#define WIN32_LEAN_AND_MEAN     // Exclude rarely-used stuff from Windows headers
#include "plugin.hpp"
#include <shellapi.h>
#include <shlobj.h>
#include <malloc.h>


enum {DoCopy, DoCut, DoNothing, DoPaste, DoLink};

enum {
  MCopy, MCut, MTitle, MPaste, MLink,
  MError, MOk,
  MErrParseDisplayNameFile, MErrSetCutAttributes,
  MErrSetClipboard, MErrLinkCreate, MErrPrepareToCopy
};

static const char szUsingThePlugin[] = "Use";
static const char szLnk           [] = ".lnk";

struct PluginStartupInfo Info;
HANDLE hHeap;

static HRESULT CreateLink(LPCSTR lpszPathObj, LPSTR lpszPathLink/*{, LPSTR lpszDesc*/);
static void ShowErrorMsg(int ErrorNumber, int ErrorCode);
static void PanelsRedraw(void);
static unsigned long __stdcall CopyThread(void *);
static const char *GetMsg(int MsgId);
static void parse_cmdline(char *cmdstart, char **argv, char *args, UINT *numargs, UINT *numchars);

static __inline int IsSpace(int c) { return ((c) == 0x20 || ((c) >= 0x09 && (c) <= 0x0D)); }

static void ShowErrorMsg(int ErrorNumber, int ErrorCode)
{
  char lpBuffer[MAX_PATH];
  wsprintf(lpBuffer, GetMsg(ErrorNumber), ErrorCode);
  const char *MsgItems[] = {GetMsg(MError), lpBuffer, GetMsg(MOk)};
  Info.Message(Info.ModuleNumber, FMSG_WARNING, szUsingThePlugin, MsgItems,
    sizeof(MsgItems) / sizeof(MsgItems[0]), 1);
}

static __forceinline HRESULT CreateLink(LPCSTR lpszPathObj, LPSTR lpszPathLink/*{, LPSTR lpszDesc*/)
{
  HRESULT hres;
  IShellLink* psl;

  // Get a pointer to the IShellLink interface.
  hres = CoCreateInstance(CLSID_ShellLink, NULL,
    CLSCTX_INPROC_SERVER, IID_IShellLink, (void**)&psl);
  if (SUCCEEDED(hres)) {
    IPersistFile* ppf;

    // Set the path to the shortcut target, and add the
    // description.
    psl->SetPath(lpszPathObj);

//    psl->SetDescription(lpszDesc);

    // Query IShellLink for the IPersistFile interface for saving the
    // shortcut in persistent storage.
    hres = psl->QueryInterface(IID_IPersistFile, (void**)&ppf);

    if (SUCCEEDED(hres)) {
      WORD wsz[MAX_PATH];

      // Ensure that the string is ANSI.
      MultiByteToWideChar(CP_ACP, 0, lpszPathLink, -1, wsz, MAX_PATH);

      // Save the link by calling IPersistFile::Save.
      hres = ppf->Save(wsz, TRUE);
      ppf->Release();
    }
    psl->Release();
  }
  return hres;
}

static __inline void PanelsRedraw(void)
{
  SetFileApisToOEM();
  Info.Control(INVALID_HANDLE_VALUE, FCTL_UPDATEPANEL, NULL);
  Info.Control(INVALID_HANDLE_VALUE, FCTL_REDRAWPANEL, NULL);
}

static unsigned long __stdcall CopyThread(void *pci)
{
  SHFILEOPSTRUCT FOS;
  DWORD dwStart = GetTickCount();

  memset(&FOS, 0, sizeof(FOS));
  FOS.pTo = (LPCTSTR)pci + sizeof(DWORD);
  if (*(DWORD*)pci == DROPEFFECT_MOVE) FOS.wFunc = FO_MOVE; else FOS.wFunc = FO_COPY;
  FOS.pFrom = &FOS.pTo[lstrlen(FOS.pTo) + 2];
  FOS.fFlags = FOF_ALLOWUNDO;
  SHFileOperation(&FOS);
  HeapFree(hHeap, 0, pci);
  if (GetTickCount() < (dwStart + 2000)) PanelsRedraw();
  return TRUE;
}

static const char *GetMsg(int MsgId)
{
  return (Info.GetMsg(Info.ModuleNumber, MsgId));
}

static void parse_cmdline(char *cmdstart, char **argv, char *args, UINT *numargs, UINT *numchars)
{
  char *p;
  int inquote;                    /* 1 = inside quotes */
  int copychar;                   /* 1 = copy char to *args */
  unsigned numslash;              /* num of backslashes seen */

  *numchars = 0;
  *numargs = 0;

  p = cmdstart;

  inquote = 0;

  /* loop on each argument */
  for(;;) {

    while (IsSpace((int)*p)) ++p;

    if (*p == '\0')break;   /* end of args */

    /* scan an argument */
    if (argv) *argv++ = args;     /* store ptr to arg */
    ++*numargs;

    /* loop through scanning one argument */
    for (;;) {
      copychar = 1;
      /* Rules: 2N backslashes + " ==> N backslashes and begin/end quote
         2N+1 backslashes + " ==> N backslashes + literal "
         N backslashes ==> N backslashes */
      numslash = 0;
      while (*p == '\\') { /* count number of backslashes for use below */
        ++p;
        ++numslash;
      }
      if (*p == '\"') {
      /* if 2N backslashes before, start/end quote, otherwise
        copy literally */
        if ((numslash & 1) == 0) {
          if (inquote) {
            if (p[1] == '\"')
              p++;    /* Double quote inside quoted string */
            else        /* skip first quote char and copy second */
              copychar = 0;
          } else copychar = 0;       /* don't copy quote */
          inquote = !inquote;
        }
        numslash >>= 1;             /* divide numslash by two */
      }

      /* copy slashes */
      while (numslash--) {
        if (args) *args++ = '\\';
        ++*numchars;
      }

      /* if at end of arg, break loop */
      if (*p == '\0' || (!inquote && IsSpace((int)*p))) break;

      /* copy character into argument */
      if (copychar) {
        if (args) *args++ = *p;
        ++*numchars;
      }
      ++p;
    }

    /* null-terminate the argument */

    if (args) *args++ = '\0';          /* terminate string */
    ++*numchars;
  }
}

HANDLE WINAPI _export OpenPlugin(int OpenFrom, int Item)
{
  static UINT CF_PREFERREDDROPEFFECT = RegisterClipboardFormat(CFSTR_PREFERREDDROPEFFECT);
  IDataObject *pData;
  IShellFolder *pDesktop = NULL, *pFolder = NULL;
  FORMATETC form = {0, NULL, DVASPECT_CONTENT, -1, TYMED_HGLOBAL};
  STGMEDIUM medium;
  HRESULT hres = NOERROR;
  char szPath[MAX_PATH];
  char szName[MAX_PATH];

  UINT size, count, i;
  char *pf;
  LPMALLOC pMalloc = NULL;
  LPITEMIDLIST *ppidl;
  wchar_t wPath[MAX_PATH];
  wchar_t wName[MAX_PATH];
  void *p;
  DWORD Dummy;
  struct PanelInfo PInfo;
  DWORD effect;

  int ExitCode = DoNothing;
  char **SelectedFiles;
  int SelectedFilesCount;

  if (OpenFrom == OPEN_COMMANDLINE) {
    parse_cmdline((char*)Item, NULL, NULL, (UINT*)&SelectedFilesCount, &i);
    SelectedFiles = (char**)alloca(SelectedFilesCount * sizeof(char*) + i * sizeof(char) + 1);
    parse_cmdline((char*)Item, SelectedFiles, ((char*)SelectedFiles) + sizeof(char*) * SelectedFilesCount,
      (UINT*)&SelectedFilesCount, &i);
    SelectedFilesCount--;
    static const char CmdParams[DoLink + 1][6] = {"copy", "cut", "", "paste", "link"};
    for (i = DoCopy; i <= DoLink; i++)
      if (lstrcmpi(SelectedFiles[0], CmdParams[i]) == 0) ExitCode = i;
    SelectedFiles = &SelectedFiles[1];
  }
  if (ExitCode == DoNothing) {
    struct FarMenuItem MenuItems[5];

    OpenFrom=OPEN_PLUGINSMENU;
    memset(MenuItems, 0, sizeof(MenuItems));
    MenuItems[DoNothing].Separator = 1;
    for (i = DoCopy; i <= DoLink; i++) lstrcpy(MenuItems[i].Text, GetMsg(i));

    ExitCode = Info.Menu(Info.ModuleNumber, -1, -1, 0, FMENU_WRAPMODE,
      GetMsg(MTitle), NULL, szUsingThePlugin, NULL, NULL, MenuItems, 5);
    Info.Control(INVALID_HANDLE_VALUE, FCTL_GETPANELINFO, &PInfo);
    SelectedFilesCount = PInfo.SelectedItemsNumber;
  }

  SetFileApisToANSI();
  switch (ExitCode) {
    case DoCopy:
    case DoCut: {
      OleSetClipboard(NULL);
      ppidl = (LPITEMIDLIST*)alloca(sizeof(LPITEMIDLIST) * SelectedFilesCount);
      if (ppidl) {
        OleInitialize(NULL);
        SHGetMalloc(&pMalloc);
        SHGetDesktopFolder(&pDesktop);

        count = 0;
        for (i = 0; (int)i < SelectedFilesCount; i++) {
          if (OpenFrom == OPEN_COMMANDLINE)
            lstrcpyn(szName, SelectedFiles[i], sizeof(szName)/sizeof(szName[0]));
          else
            lstrcpy(szName, PInfo.SelectedItems[i].FindData.cFileName);
          OemToChar(szName, szName);
          if (!GetFullPathName(szName, sizeof(szPath), szPath, &pf)) continue;
          if (lstrcpy(szName, pf)) {
            *pf = '\0';
            MultiByteToWideChar(CP_ACP, 0, szName, -1, wName, sizeof(wName) / sizeof(wName[0]));
            if (!pFolder) {
              MultiByteToWideChar(CP_ACP, 0, szPath, -1, wPath, sizeof(wPath) / sizeof(wPath[0]));
              hres = pDesktop->ParseDisplayName(0, 0, wPath, 0, ppidl, 0);
              if FAILED(hres) break;
              hres = pDesktop->BindToObject(ppidl[0], 0, IID_IShellFolder, (void**)&pFolder);
              pMalloc->Free(ppidl[0]);
              if FAILED(hres) break;
            }
            if SUCCEEDED(pFolder->ParseDisplayName(0, 0, wName, 0, &ppidl[count], 0)) count++;
          }
        }

        if FAILED(hres) ShowErrorMsg(MErrPrepareToCopy, hres);
        if (count != i) ShowErrorMsg(MErrParseDisplayNameFile, i - count);

        if (count) {
          hres = pFolder->GetUIObjectOf(0, count, (LPCITEMIDLIST*)ppidl, IID_IDataObject, 0, (void**)&pData);
          for (i = 0; i < count; i++) pMalloc->Free(ppidl[i]);
          if SUCCEEDED(hres) {
            if (ExitCode == DoCut) {
              form.cfFormat = (WORD)CF_PREFERREDDROPEFFECT;
              medium.tymed = TYMED_HGLOBAL;
              medium.pUnkForRelease = NULL;
              medium.hGlobal = GlobalAlloc(GPTR, sizeof(effect));
              if (medium.hGlobal) {
                *(DWORD*)medium.hGlobal = DROPEFFECT_MOVE;
                hres = pData->SetData(&form, &medium, TRUE);
                if FAILED(hres) {
                  ReleaseStgMedium(&medium);
                  ShowErrorMsg(MErrSetCutAttributes, hres);
                }
              }
            }
            hres = OleSetClipboard(pData);
            if SUCCEEDED(hres) hres = OleFlushClipboard();
            pData->Release();
          }
        }
        if FAILED(hres) ShowErrorMsg(MErrSetClipboard, hres);

        if (pFolder)  pFolder->Release();
        if (pDesktop) pDesktop->Release();
        if (pMalloc) pMalloc->Release();
        OleUninitialize();
        PanelsRedraw();
        break;
      }
    }
    case DoPaste:
    case DoLink: {
      OleInitialize(NULL);
      hres = OleGetClipboard(&pData);
      if SUCCEEDED(hres) {
        effect = DROPEFFECT_COPY;
        form.cfFormat = (WORD)CF_PREFERREDDROPEFFECT;
        hres = pData->GetData(&form, &medium);
        if SUCCEEDED(hres) {
          p = GlobalLock(medium.hGlobal);
          if (p) {
            if ((*(DWORD*)p & DROPEFFECT_MOVE) == DROPEFFECT_MOVE) effect = DROPEFFECT_MOVE;
            GlobalUnlock(p);
          }
          ReleaseStgMedium(&medium);
        }
        form.cfFormat = CF_HDROP;
        hres = pData->GetData(&form, &medium);
        if SUCCEEDED(hres) {
          DWORD *ci;

          size = 0;
          count = DragQueryFile((HDROP)medium.hGlobal, 0xFFFFFFFF, NULL, 0);
          if (ExitCode == DoPaste) {

            ci = (DWORD*)HeapAlloc(hHeap, HEAP_GENERATE_EXCEPTIONS | HEAP_ZERO_MEMORY,
              (count + 1) * (MAX_PATH + 1) + sizeof(DWORD) + 2);
            size += GetCurrentDirectory(MAX_PATH,(LPTSTR)&ci[1]) + sizeof(DWORD) + 2;
          }
          for (i = 0; i < count; i++) {
            if (ExitCode == DoPaste) {
              size += DragQueryFile((HDROP)medium.hGlobal, i, &((char*)ci)[size], MAX_PATH) + 1;
            } else {
              DragQueryFile((HDROP)medium.hGlobal, i, szName, sizeof(szName));
              if (GetFullPathName(szName, sizeof(szPath), szPath, &pf) != 0) {
                int j, cnt = 1;
                for (j = lstrlen(pf) - 1; j >= 0 && pf[j] != '.'; j--);
                if (j < 0) j = lstrlen(pf);
                lstrcpy(&pf[j], szLnk);
                WIN32_FIND_DATA fd;
                HANDLE hFind;
                while (hFind = FindFirstFile(pf, &fd), hFind != INVALID_HANDLE_VALUE ) {
                  FindClose(hFind);
                  wsprintf(&pf[j], " (%d)%s", ++cnt, szLnk);
                }
                hres = CreateLink(szName, pf/*, NULL*/);
              }
              if FAILED(hres) ShowErrorMsg(MErrLinkCreate, hres);
            }
          }
          ReleaseStgMedium(&medium);
          pData->Release();
          if (ExitCode == DoPaste) {
            if (effect == DROPEFFECT_MOVE) OleSetClipboard(NULL);
            *ci = effect;
            if( int(GetVersion()) < 0 ) { // W9x
              CharToOemBuff((LPTSTR)&ci[1], (LPTSTR)&ci[1], (count + 1) * (MAX_PATH + 1) + 2);
              SetFileApisToOEM();
            }
            CreateThread(NULL, 0, CopyThread, ci, 0, &Dummy);
          } else PanelsRedraw();
        }
      }
      OleUninitialize();
      break;
    }
  }
  SetFileApisToOEM();
  return INVALID_HANDLE_VALUE;
}

void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  static const char *PluginMenuStrings[1];
  PluginMenuStrings[0] = GetMsg(MTitle);
  Info->PluginMenuStrings = PluginMenuStrings;
  Info->StructSize = sizeof(*Info);
  Info->PluginMenuStringsNumber = sizeof(PluginMenuStrings) / sizeof(PluginMenuStrings[0]);
  Info->CommandPrefix="fclip";
}

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *pInfo)
{
  Info = *pInfo;
  hHeap = GetProcessHeap();
}
